#! /usr/bin/perl 

# Copyright (C) 2006 Paul Thomas.
# This file may be copied and redistributed freely, without advance permission,
# provided that this Copyright statement is reproduced with each copy.

# LIMITATION OF WARRANTY
# NOTHING IN THIS AGREEMENT WILL BE CONSTRUED AS A REPRESENTATION MADE OR
# WARRANTY GIVEN BY PAUL THOMAS OR ANY THIRD PARTY THAT THE USE OF
# DATA PROVIDED HEREUNDER WILL NOT INFRINGE ANY PATENT, COPYRIGHT, TRADEMARK
# OR OTHER RIGHTS OF ANY THIRD PARTY. DATA IS PROVIDED "AS IS" WITHOUT
# WARRANTY OF ANY KIND WHATSOEVER, EXPRESS OR IMPLIED, INCLUDING IMPLIED
# WARRANTIES OF MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE. PAUL
# THOMAS MAKES NO WARRANTY THAT ITS DATA DOES NOT CONTAIN ERRORS.

#####
#####
##
## conserved2profile.pl
##
#####
#####


# necessary libraries
use Math::BigInt;
use strict;
use FastaFile;

# default values
my $wild_aa="";
my $sf_name="none";
my %aa2num=qw(A 0 C 1 D 2 E 3 F 4 G 5 H 6 I 7 K 8 L 9 M 10 N 11 P 12 Q 13 R 14 S 15 T 16 V 17 W 18 Y 19);
my ($SF_4_leaf_GLOB, $leaf_4_SF_GLOB, $ids4SFs_GLOB, $seq_GLOB, $SFs_GLOB);
my ($pir_file,$col,$tree_file,$sf_name,$out_file,$wild_aa,$mut_aa,$out_seqsUser,$wtFileUser,$adjWts,$do_SF,$priorFile,$errFile,$do_compact_evolSF, $sfan_file);
my $sfConservedAA;
my $tmpdirPath = "/tmp";

# get command-line arguments
use Getopt::Long;
$Getopt::Long::ignorecase=0; #case sensitive
&GetOptions 
(
 "h" => sub {&usage()},        # -h for help
 "o=s" => \$out_file,          # -o for (o)utput file (redirect STDOUT)
 "S"   => \$do_SF,             # -S for (S)F profile file 
 "t=s" => \$tree_file,         # -t for (t)ree file
 "T=s" => \$tmpdirPath,     # -T for path to (T)mp directory
 "p=s" => \$pir_file,          # -p for (p)ir file
 "f=s" => \$sfan_file,         # -f for s(f)an file
 "c=i" => \$col,               # -c for (c)olumn
 "s=s" => \$sf_name,           # -s for (s)f name
 "w=s" => \$wild_aa,           # -w for (w)ild type AA
 "m=s" => \$mut_aa,            # -m for (m)utant type AA
 "A"   => \$adjWts,            # -a for (A)adjusted Weights
 "P=s"       => \$priorFile,   # -P for (P)rior file
 "outSeqs=s" => \$out_seqsUser,# - outSeqs   for printing out selected seqs
 "outWts=s"  => \$wtFileUser,  # - outWts:s  for printing out weights
 "e=s" => \$errFile,           # -e for (e)rror file (redirect STDERR)
 "compact" => \$do_compact_evolSF # -compact for compact evol subfam  
 
);

&usage("Wild type aa $wild_aa is not valid") unless (defined($aa2num{$wild_aa}));
&usage("Mutant type aa $mut_aa is not valid") unless (defined($aa2num{$mut_aa}));
&usage("tmp directory $tmpdirPath cannot be found\n") unless (-d $tmpdirPath);


my @identical = @ARGV if (@ARGV);

#######################################################
###  MAIN BEGINS HERE
#######################################################

#set stuff based on user options, and set up temporary files
my ($out_seqs);
if ($out_seqsUser) {
  $out_seqs = $out_seqsUser;
} else {
  $out_seqs="$tmpdirPath/out.seqs.$$";
}
my ($wtFile);
if ($wtFileUser) {
  $wtFile = $wtFileUser;
} else {
  $wtFile ="$tmpdirPath/wts.$$";
}

#if the subfamily is not defined, this means we are doing a GHMM analysis
if(! $sf_name){
  my $alignFile;
  if (@identical > 0) {
    #remove identical from pir file
    my $newAlnSeqs = &removeIdenticalFromPirFile($pir_file,\@identical);
    $alignFile = "$tmpdirPath/consv2prof.$$.pir";
    open(FH,">$alignFile");
    print FH $newAlnSeqs;
    close(FH);
  } else {
    $alignFile = $pir_file;
  }

  my $cmd_dirichlet_CF="alignment2dirichlet.pl -a $alignFile -w $wild_aa -m $mut_aa -c $col -o $out_file -C"; ## clean_alignment
  $cmd_dirichlet_CF.= " -W $wtFile" if ($wtFileUser);
  $cmd_dirichlet_CF.= " -A $adjWts" if ($adjWts);
  $cmd_dirichlet_CF.= " -P $priorFile" if ($priorFile);
  `$cmd_dirichlet_CF`;
  unlink ($alignFile) if (@identical > 0);

  unlink $out_seqs  unless ($out_seqsUser);
  unlink $wtFile    unless ($wtFileUser);

  exit(0);
}

#do subfamily or evolutionary subfamily analysis
&read_PIR_file($pir_file,\@identical);


#check that there is at least one sequence ID in the SF to be analyzed
my @ids = $ids4SFs_GLOB->{$sf_name};
die "Error - died: there are no sequences for $sf_name\n" if ( (! @ids) || (@ids == 0));


my $conserved_SFs;
#force it do SF, and not do evolutionary SF if user requests so
if ($do_SF) {
  my $sfList = ();
  $sfList->[0] = $sf_name;
  &print_selected_SFs($out_seqs,$sfList);
} else {
  #determine the SFs that make upt he evol SF
  $conserved_SFs = &tree2conserved_SFs(0,$seq_GLOB,$ids4SFs_GLOB,$col,$sf_name);
  if(! defined($conserved_SFs)){ ## SF is not conserved
    my $sfList = ();
    $sfList->[0] = $sf_name;
    &print_selected_SFs($out_seqs,$sfList);
  } else{ ## do eSF analysis
    &print_selected_SFs($out_seqs,$conserved_SFs);
  }
}

my $cmd_dirichlet="alignment2dirichlet.pl -a $out_seqs -w $wild_aa -m $mut_aa -c $col -o $out_file "; ## alignment file is clean
$cmd_dirichlet .= " -W $wtFile" if ($wtFileUser);
$cmd_dirichlet .= " -A $adjWts" if ($adjWts);
$cmd_dirichlet .= " -P $priorFile" if ($priorFile);
`$cmd_dirichlet`;

#add conserved SFs that make up evol SF 
&append_SF_names2out_file($out_file,$conserved_SFs);

unlink $out_seqs  unless ($out_seqsUser);
unlink $wtFile    unless ($wtFileUser);

#######################################################
###  SUBROUTINES BEGIN HERE
#######################################################

sub removeIdenticalFromPirFile {
  my $pir_file = shift;
  my $identical = shift;

  my $ff = new FastaFile($pir_file);
  $ff->open();
  my $newAlnSeqs;
  while (my $seq = $ff->nextSeq()) {
    my ($id) = FastaFile::id($seq);
    next if ($id=~/Subfamily/);
    next if (grep(/^$id$/,@$identical));
    $newAlnSeqs .= $seq,"\n";
  }
  $ff->close();
  return $newAlnSeqs;
}

#store seq info in $seq_GLOB, list of SF in $SFs_GLOB, and IDs in $ids4Sfs_GLOB
sub read_PIR_file(){
  my $pir_file = shift;
  my $identical = shift;
  
  my $ff = new FastaFile($pir_file);
  $ff->open();
  my $SF_id;
  while (my $seq = $ff->nextSeq()) {
    my ($id) = FastaFile::id($seq);
    next if (grep(/^$id$/,@$identical)); #remove identical
    if ($id=~/^Subfamily:(SF\d+)/) {
      $SF_id = $1;
      $SFs_GLOB->{$SF_id}=1;
    } else {
      push(@{$ids4SFs_GLOB->{$SF_id}},$id);
      my ($seqText,$title) = FastaFile::splitFasta($seq,1);
      $seqText=~s/[\cM]+//g;
      $seqText=~s/\n//g;
      $seqText=~s/[a-z]//g;
      $seqText=~s/\.//g;
      $seq_GLOB->{$id} .=$seqText;
    }
  }
  $ff->close();
}


####
#
# check if subfamily is conserved.  Ignore Xs and dashes#
# return 0 not conserved, 1 when conserved; 2 when empty
# all X, return conserved, since this never should happen
#
# if all sequences have been removed from SF (@identical), the SF will be 
# treated just like it has all gaps
####
sub is_SF_conserved(){
  my $sf_name = shift;
  my $seq_GLOB = shift;
  my $ids4SFs = shift;
  my $col = shift;
  my $specifiedAA = shift;  

  #loop through each ID of the SF
  my $conservedAA;
  my $all_gapsOrX=1;
  foreach my $id(@{$ids4SFs->{$sf_name}}){
    #get the amino acid for column of interest
    my $curr_aa=substr($seq_GLOB->{$id},$col-1,1);

    ## ignore gaps and Xs
    if ($curr_aa eq "X" || $curr_aa eq "-") {
      next;
    } else {
      $all_gapsOrX =0;
    }
    
    if ($specifiedAA) {
      return 0 if ($curr_aa ne $specifiedAA);
    } else {
      return 0 if ($conservedAA && ($curr_aa ne $conservedAA));
    }
    $conservedAA = $curr_aa;
  }

  die "All X/gaps in the starting SF $sf_name at $col\n" if ($all_gapsOrX && (! $specifiedAA));
      

  if ($all_gapsOrX) {
    return (2,$conservedAA);
  } else {
    return (1,$conservedAA);
  }
}


sub node2conserved_SFs(){
  my ($node,$seq_GLOB,$ids4SFs,$col,$specifiedAA)=@_;
  
  my $current_node=$node;
  my $first_time=1;  
  my $conserved_SFs="";
  my $is_checked;
  my $all_sfs_empty=1;
  
  while($current_node ne $node || $first_time){
    $first_time=0;
    ### print $current_node,"\n";
    if(defined(my $SF=$SF_4_leaf_GLOB->{$current_node})){
      my ($is_sf_conserved,$tmpAA)=&is_SF_conserved($SF,$seq_GLOB,$ids4SFs,$col,$specifiedAA); ## 1-consrvd not empty; 2- cempty
      if($is_sf_conserved) {
	my $sf=$SF;
	$sf=$SF."*" if ($is_sf_conserved ==2);
	$all_sfs_empty=0 if ($is_sf_conserved ==1);
	# push(@{$conserved_SFs},$sf);
	$conserved_SFs.="$sf ";
	last if ($current_node eq $node); ## if ORIG $node is terminal
	$current_node= &traverse_next_node($current_node,$is_checked); ## right sibling if left
	## parent
      }
      else{
	$conserved_SFs=undef; ### make array empty;
	last;
      }
    }
    elsif($is_checked->{$current_node}){
      $current_node= &traverse_next_node($current_node,$is_checked); ###
    }
    else{
      $current_node= &left_child($current_node);
    }
  }
  
  $conserved_SFs.="!" if (defined($conserved_SFs) &&  $all_sfs_empty);
  return $conserved_SFs;
}

#create evolutionary SF
sub node2enhanced_conserved_SFs(){
  my ($node,$sf_name,$seq_GLOB,$ids4SFs,$col,$specifiedAA)=@_;

  ### traverse a tree up
  my $current_node=$node;
  my $next_node;
  my $SFs;
  my $SFs_next;
  
  push(@{$SFs},$sf_name);
  return $SFs if ($current_node eq "1");

  while(defined($SFs_next=&node2conserved_SFs(&sibling($current_node),$seq_GLOB,$ids4SFs,$col,$specifiedAA))){
    push(@{$SFs},$SFs_next);
    $current_node=&parent($current_node);
    last if ($current_node eq "1");
  }
  return $SFs;
}

sub tree2conserved_SFs(){
  my $is_SFonly  = shift;
  my $seq_GLOB = shift;
  my $ids4SFs = shift;
  my $col = shift;
  my $sf_name = shift;
  
  my $parent;
  my $current_SF;
  my $next_node;

  #see if SF is conserved
  my ($is_sf_conserved,$conservedAA) = &is_SF_conserved($sf_name,$seq_GLOB,$ids4SFs,$col);
  $sfConservedAA = $conservedAA if ($is_sf_conserved);

  my $conserved_SFs = "$sf_name" if ($is_sf_conserved);

  #if only doing SF, or the SF is not conserved you are done
  #since don't need to calculate EVOL SF
  return $conserved_SFs if ($is_SFonly || (! $is_sf_conserved));

  #if SF is conserved, and not just doing SF, then read in tree structure
  #in order to evol SF
  &read_tree($tree_file);

  #get node info
  my $current_node = $leaf_4_SF_GLOB->{$sf_name};


  #now look to see whether sibling SFs have the same conserved amino acid
  #so that we can created the extended/evolutionary SF
  #the original SF is always retained in the $conserved_SFs column
  $conserved_SFs = &node2enhanced_conserved_SFs($current_node,$sf_name,$seq_GLOB,$ids4SFs,$col,$conservedAA);

  if($do_compact_evolSF){## GET RID OF FLANKING NODES THAT HAVE ALL GAPS IN ALIGNMENT
    while(my $sfs=pop(@{$conserved_SFs})){
      if ($sfs!~/\!$/){
	push(@{$conserved_SFs},$sfs); ## write ot back
	last;
      }
    }
  }
  return  $conserved_SFs;
}



sub get_tree_structure(){
  my $path_GLOB=shift;
  my $line=shift;
  my $sfan_file = shift;

# This part is added to the script to handle the change of the tree format from 
# v.6.1 to v.7.0

  my %sfan;
  open (SA, $sfan_file);
  while (my $line=<SA>){
    chomp $line;
    
    my ($sf, $an)=split(/\t/, $line);  
    $sfan{$an}=$sf;
    }
  close (SA);

   
  my $SF_id;
  
  # print "LINE=",$line;
  chomp $line;

  $line=~s/\[\&\&NHX\:\w+\=\d+\>\d+\:S\=\w+\:ID\=AN\d+\]//g;
  $line=~s/\[\&\&NHX\:\w+\=\d+\>\d+\:S\=\w+\-\w+\:ID\=AN\d+\]//g;
  $line=~s/\[\&\&NHX\:\w+\=\d+\>\d+\:ID\=AN\d+\]//g;
  $line=~s/[\;\cM]//g;
  $line =~s/\:-*(\d)+(\.(\d)+)?//g;

  my @array=split (/([\(\)\,])/, $line);

  my @array1;
  foreach my $item (@array){

    next unless ($item);

    if (defined $sfan{$item}){
        push (@array1, $sfan{$item});
    }else{
        push (@array1, $item);
    }
  }
  close (TR);

  my $line1=join("", @array1);

  $line1=~s/AN\d+//g;
  $line1=~s/\,//g;

  while ($line1=~/\(\)|\(SF\d+\)/){
    $line1=~s/\(\)//g;
    $line1=~s/\((SF\d+)\)/$1/g;
  }

  while ($line1=~/SF\d+SF\d+/){
    
    $line1=~s/(SF\d+)(SF\d+)/$1\,$2/g;
  }

# This ends the added section
  
  my @splits=split (/([\(\)\,])/, $line1);
  
  foreach my $split(@splits){
     if($split eq "("){
       $path_GLOB=&left_child($path_GLOB);
     } 
     elsif( $split eq ","){
       $path_GLOB=&right_sibling($path_GLOB);  ### handles trifurcation . ex: 1102
       print $line,"\n" if($path_GLOB eq "12");
     } 
     elsif( $split eq ")"){
       $path_GLOB=&parent($path_GLOB);
       
     }
     elsif($split =~ /(\w+)/){
       $SF_id=$1;
       if($SF_id=~/^SF\d+/){ 
	 $SF_4_leaf_GLOB->{$path_GLOB}=$SF_id;
	 $leaf_4_SF_GLOB->{$SF_id}=$path_GLOB;
       }
       else{
	 die "Something is wrong";
       }
     }
   }
}
  

sub read_tree(){  
  my $tree_file=shift;
  
  my $path_GLOB=Math::BigInt->new("1");
  
  open TREE, "<$tree_file" or die "Can't open tree: $tree_file";
  
  while(<TREE>){
    next if /^SF\d+\:/; ## read only the first line
    &get_tree_structure($path_GLOB, $_, $sfan_file);
  }
}

sub left_child(){
  my $path=shift;
    
  my $path_l=Math::BigInt->new($path);
  
  return 10*$path_l;
}

sub right_child(){
  my $path=shift;
  my $path_l=Math::BigInt->new($path);
  
  return 10*$path_l+1; 
}


sub right_sibling(){
  my $path=shift;
  my $path_l=Math::BigInt->new($path);

  my $path_out=$path_l+1;
  
  my $result=$path_l->bmod("10");
  if($result ==1){
    my $foo=0;
  }
  die " **** NOT A BIFURCATING TREE :",  $path,"\t",$path_out, "\n" if ($result == 1);
  # $is_unrooted_GLOB=1 if ($result == 1);
  
  return $path_out;
}

sub left_sibling(){
  my $path=shift;

  my $path_l=Math::BigInt->new($path);
  return $path_l-1;  ## take left sibling
}

sub sibling(){
  my $path=shift;

  my $path_l=Math::BigInt->new($path);
  my $path_tmp=Math::BigInt->new($path);
  my $result=$path_tmp->bmod("10");  

  if($result eq "1"){
    return $path_l-1;  ## take left sibling
  }
  else{
    return $path_l+1; ## take right sibling
  }
}

sub parent(){
  my $path=shift;
  my $path_l=Math::BigInt->new($path);
  my $result=$path_l->bdiv("10");
  # print "path= ",$path,"   Path_l= ",$path_l,"   result= ",$result,"\n";
  
  return $result;
}

sub traverse_next_node(){
  my $path=shift;
  my $is_checked=shift;
  
  my $path_l=Math::BigInt->new($path);
  
  
  if($path_l->bmod(10) ){
    my $parent=&parent($path);
    $is_checked->{$parent}=1;
    return $parent;
  } else{
    return  &right_sibling($path);
  }
}

sub are_siblings(){ ## bifurcating version 
  my $n1=shift;
  my $n2=shift;
  
  my $diff=(Math::BigInt->new($n1)-Math::BigInt->new($n2))->babs;  
  return $diff->bcmp("1")?0:1;
}

sub place_in_sorted() {
  my $sorted_array=shift;
  my $new=shift;
  
  my $new_l=Math::BigInt->new($new);
  my $temp;
  
  my $i=scalar(@{$sorted_array});
  $sorted_array->[$i]=Math::BigInt->new($new);
  while($sorted_array->[$i]->bcmp($sorted_array->[$i-1]) == -1 && $i>0 ){
    $temp=$sorted_array->[$i-1];
    $sorted_array->[$i-1]=$sorted_array->[$i];
    $sorted_array->[$i]=$temp;
    $i--;
  }
  return $sorted_array;
}

sub print_selected_SFs(){
  my $out_seqs_file=shift;
  my $SF_list=shift;
  
  open FILE,">$out_seqs_file" or die "Can't open $out_seqs_file";
  my $numIds = 0;
  foreach my $sfs_string (@{$SF_list}){
    $sfs_string=~s/\!//g; ## clean from !
    foreach my $sf(split /\s/,$sfs_string){
      my ($sf_clean)=$sf=~/(SF\d+)/; ## clean out "*" at the end for empty SFs
      delete $SFs_GLOB->{$sf_clean}; ## remove SF from the list of all SFs
      my @ids = $ids4SFs_GLOB->{$sf_clean};
      foreach my $id(@{$ids4SFs_GLOB->{$sf_clean}}){
	print FILE ">$id\n";
	print FILE "$seq_GLOB->{$id}\n";
	$numIds++;
      }
    }
  }
  close(FILE);
  if ($numIds == 0) {
    print STDERR "Error/warning: there are zero sequence IDs to analyze\n";
  }
}

sub append_SF_names2out_file{
  
  my ($out_file,$conserved_eSFs)=@_;
  
  open OUT_FILE, ">>$out_file" or die "Can't append 2to $out_file";
  
  print OUT_FILE "\nConserved SFs=";
  my $sfString;
  foreach my $sf(@{$conserved_eSFs}){
    $sfString .= "$sf,";
  }
  $sfString=~s/,$//;
  print OUT_FILE "$sfString\n";
  ## check whether all CF is conserved
  print OUT_FILE "Evolutionary family=";
  if(scalar(keys(%{$SFs_GLOB}))){
    print OUT_FILE "NO\n";
  }
  else{
    print OUT_FILE "YES\n";
  }

  if ($sfConservedAA) {
    print OUT_FILE "\nconserved amino acid in SF=$sfConservedAA\n";
  }
  
  close(OUT_FILE);
}


####
# sub usage - displays help to user, showing available options
####
sub usage {
    my $error = shift;

    print "Error: $error\n\n" if ($error);

    print <<__EOT;

myprog.pl - a program to do something

Usage:
myprog.pl {args} 

Where args are:
\t-h for help (this message)
\t-i (i)nput file (redirect STDIN)
\t-o (o)utput file (redirect STDOUT)
\t-e (e)rror file (redirect STDERR)
\t-v (v)erbose (debug info to STDERR)
\t-V (V)ery verbose (debug info to STDERR)

__EOT

exit(-1);
}


