/*
Copyright (C) 2006 Paul Thomas
This file may be copied and redistributed freely, without advance permission,
provided that this Copyright statement is reproduced with each copy.

LIMITATION OF WARRANTY
NOTHING IN THIS AGREEMENT WILL BE CONSTRUED AS A REPRESENTATION MADE OR
WARRANTY GIVEN BY PAUL THOMAS OR ANY THIRD PARTY THAT THE USE OF
DATA PROVIDED HEREUNDER WILL NOT INFRINGE ANY PATENT, COPYRIGHT, TRADEMARK
OR OTHER RIGHTS OF ANY THIRD PARTY. DATA IS PROVIDED "AS IS" WITHOUT
WARRANTY OF ANY KIND WHATSOEVER, EXPRESS OR IMPLIED, INCLUDING IMPLIED
WARRANTIES OF MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE. PAUL
THOMAS MAKES NO WARRANTY THAT ITS DATA DOES NOT CONTAIN ERRORS.
*/


#include "gen_incl.h"
#include "fasta_seq.h"

/* reads a sequence (in simple "fasta" format) from a file, and
fills up a structure (seq) */
/* note: allocates memory for fasta_seq **seq that must be freed */
/* returns EOF if EOF is reached, 1 otherwise */
int read_fasta_seq(FILE *fp, fasta_seq **seq, int flag) {
   int seqlen=0;
   char c;
   char line[MAXLINELEN];
   char tseq[MAXSEQLEN];
   fasta_seq *s;

/* read the length, and allocate memory */
   if((c=fgetc(fp))==EOF) return(EOF);
   else if(c!='>') { printf("Error: no header in fasta file\n"); return(EOF); }
   s = (fasta_seq *)malloc(sizeof(fasta_seq));
   fscanf(fp,"%[^\n]",s->comment);
   while((c=fgetc(fp))!=EOF) {
      if(c=='>') {
         ungetc(c, fp);
         tseq[seqlen] = '\0';
         s->seqlen = strlen(tseq);
         s->seq = (char *)malloc((strlen(tseq)+1)*sizeof(char));
         strcpy(s->seq,tseq);
         (*seq) = s;
         return(1);
      }
      else if(c>='A' && c<='Z') tseq[seqlen++] = c;
      else if(flag==ALIGN) { /* ignore lower case letters, numbers and periods from prettyalign */
	 if(c=='-') tseq[seqlen++] = '-';
      }
      else if(c>='a' && c<='z') tseq[seqlen++] = toupper(c);
   }
   ungetc(c, fp);
   s->seqlen = strlen(tseq);
   s->seq = (char *)malloc((strlen(tseq)+1)*sizeof(char));
   strcpy(s->seq,tseq);
   (*seq) = s;
   return(1);
}
   
void free_fasta_seq(fasta_seq *seq) {
   free(seq->seq);
   free(seq);
   return;
}

/* prints out a fasta sequence structure, for debugging purposes */
void print_fasta_seq(FILE *fp, fasta_seq *seq) {
   int i, n=0;

   fprintf(fp,">%s\n",seq->comment);
   fprintf(fp,"sequence length = %d\n", seq->seqlen);
   for(i=0; i<seq->seqlen; ++i) {
      fprintf(fp,"%c",seq->seq[i]);
      if(++n>60) { n=0; fprintf(fp,"\n"); }
   }
   if(n) fprintf(fp,"\n");

   return;
}
